import { env } from 'bun';
import { and, eq, sql } from 'drizzle-orm';
import * as jose from 'jose';
import { drizzleDB } from './db';
import { users, userSessions } from './db/schema';

if (!env.JWT_SECRET || env.JWT_SECRET.length < 16) {
  throw new Error('JWT_SECRET must be at least 16 characters long');
}

const jwtCookie = 'pt-auth';
const alg = 'HS256';
const secret = new TextEncoder().encode(env.JWT_SECRET);

const day = 1000 * 60 * 60 * 24;
let days = Number.parseInt(env.SESSION_DURATION_IN_DAYS ?? '31', 10);
days = Number.isNaN(days) ? 31 : days;
const maxAge = days * day;

const cookieInit: Bun.CookieInit = {
  name: jwtCookie,
  path: '/',
  maxAge,
  secure: true,
  sameSite: 'strict',
  httpOnly: true,
};

const renewalTime = Math.abs(Math.floor(maxAge / 7));

type JWTPayload = { token: string; userId: string };

const getUserByEmail = drizzleDB
  .select()
  .from(users)
  .where(eq(users.email, sql.placeholder('email')))
  .limit(1)
  .prepare();

const insertSession = drizzleDB
  .insert(userSessions)
  .values({
    token: sql.placeholder('token'),
    maxAge: sql.placeholder('maxAge'),
    userId: sql.placeholder('userId'),
  })
  .prepare();

async function login(email: string, password: string): Promise<string | null> {
  const result = await getUserByEmail.execute({ email: email.toLocaleLowerCase() });

  if (result.length === 0) {
    return null;
  }

  const user = result[0];
  if (!user) {
    return null;
  }

  // verify password
  const valid = await Bun.password.verify(password, user.hash);
  if (!valid) {
    return null;
  }

  const token = Bun.randomUUIDv7('base64url');

  await insertSession.execute({ token, maxAge: Date.now() + maxAge, userId: user.id });

  const payload: JWTPayload = { token, userId: user.id };

  return new jose.SignJWT(payload)
    .setProtectedHeader({ alg })
    .setIssuedAt()
    .setIssuer('skaarup.dev')
    .setAudience('skaarup.dev')
    .setExpirationTime(`${days}d`)
    .sign(secret);
}

function loginResponse(jwt: string) {
  const cookie = new Bun.Cookie({ ...cookieInit, value: jwt });
  return new Response('Login successful', {
    headers: new Headers({
      'Set-Cookie': cookie.toString(),
    }),
  });
}

function logoutHeaders() {
  const cookie = new Bun.Cookie({ ...cookieInit, maxAge: -1 });
  return new Headers({
    'Set-Cookie': cookie.toString(),
  });
}

function logoutResponse() {
  return new Response('Logout successful', { status: 303, headers: logoutHeaders() });
}

function unAuthorizedResponse() {
  return new Response('Unauthorized', { status: 401, headers: logoutHeaders() });
}

const getUserSessionByTokenAndUserId = drizzleDB
  .select()
  .from(userSessions)
  .where(
    and(
      eq(userSessions.token, sql.placeholder('token')),
      eq(userSessions.userId, sql.placeholder('userId'))
    )
  )
  .limit(1)
  .prepare();

const deleteUserSession = drizzleDB
  .delete(userSessions)
  .where(
    and(
      eq(userSessions.token, sql.placeholder('token')),
      eq(userSessions.userId, sql.placeholder('userId'))
    )
  )
  .prepare();

const updateUserSession = drizzleDB
  .update(userSessions)
  .set({ maxAge: sql.placeholder('maxAge') as any })
  .where(
    and(
      eq(userSessions.token, sql.placeholder('token')),
      eq(userSessions.userId, sql.placeholder('userId'))
    )
  )
  .prepare();

async function verify(headers: Headers) {
  const cookieMap = new Bun.CookieMap(headers.get('cookie') || '');
  const jwtToken = cookieMap.get(jwtCookie);
  if (!jwtToken) {
    // cookie not found
    return null;
  }

  let jwtPayload: JWTPayload | null = null;
  try {
    const jwtResult = await jose.jwtVerify(jwtToken, secret, {
      algorithms: [alg],
      issuer: 'skaarup.dev',
      audience: 'skaarup.dev',
      maxTokenAge: `${days}d`,
    });
    jwtPayload = jwtResult.payload as JWTPayload;
  } catch (_) {
    // invalid token
    return null;
  }

  if (!jwtPayload || !jwtPayload.token || !jwtPayload.userId) {
    // invalid token
    return null;
  }

  const result = await getUserSessionByTokenAndUserId.execute({
    token: jwtPayload.token,
    userId: jwtPayload.userId,
  });

  const session = result[0];
  if (!session) {
    // session not found
    return null;
  }

  const now = Date.now();
  if (session.maxAge <= now) {
    // session expired
    await deleteUserSession.execute({
      token: jwtPayload.token,
      userId: jwtPayload.userId,
    });
    return null;
  }

  if (session.maxAge <= now - renewalTime) {
    // renew session
    const newMaxAge = now + maxAge;
    await updateUserSession.execute({
      token: jwtPayload.token,
      userId: jwtPayload.userId,
      maxAge: newMaxAge,
    });

    return new jose.SignJWT(jwtPayload)
      .setProtectedHeader({ alg })
      .setIssuedAt()
      .setIssuer('skaarup.dev')
      .setAudience('skaarup.dev')
      .setExpirationTime(`${days}d`)
      .sign(secret);
  }

  return undefined;
}

async function verifyResponse(jwt: string) {
  const cookie = new Bun.Cookie({ ...cookieInit, value: jwt });
  return new Response('Verify successful', {
    headers: new Headers({ 'Set-Cookie': cookie.toString() }),
  });
}

function verifyFailResponse() {
  const cookie = new Bun.Cookie({ ...cookieInit, maxAge: -1 });
  return new Response('Verify failure', {
    status: 401,
    headers: new Headers({ 'Set-Cookie': cookie.toString() }),
  });
}

export default {
  login,
  loginResponse,
  logoutResponse,
  unAuthorizedResponse,
  verify,
  verifyResponse,
  verifyFailResponse,
};
